require 'essential/resource/attr_methods'
require 'essential/resource/create'
require 'essential/resource/list'
require 'essential/resource/update'
require 'essential/resource/delete'

module Essential
  class Resource
    include AttrMethods

    class << self
      attr_reader :schema

      def request(method, url: nil, params: nil, headers: nil)
        if headers
          headers = headers.clone
          sid     = headers.delete(:sid)
          token   = headers.delete(:token)
        end

        Essential.request(method, url, sid: sid, token: token, params: params, headers: headers)
      end

      protected

      def class_name
        self.name.split('::').last
      end
    end

    attr_property :sid

    def self.url
      if self == Resource
        raise NotImplementedError, 'Resource is an abstract class'
      end
      "/v2/account/#{CGI.escape(class_name.downcase)}s"
    end

    def url
      raise ArgumentError, 'sid required' if sid.nil? || sid.empty?
      format('%s/%s', self.class.url, CGI.escape(self.sid))
    end

    def initialize(sid: nil, attrs: nil, headers: nil)
      init_from(attrs)
      @attributes['sid'] = sid if sid
      @headers = headers
      self
    end

    def fetch(headers: @headers)
      json = JSON.parse self.class.request(:get, url: self.url, headers: headers)
      init_from json
      self
    end

    def init_from(attributes)
      attributes = (attributes || {}).clone

      # reject reject keys that aren't actually our attrs
      @attributes = filter_attrs(attributes)

      if self.class.schema
        self.class.schema.each do |key,type|
          apply_schema(key, type)
        end
      end
      self
    end

    def loaded?
      @attributes.size > 1
    end

    def as_json(opts={})
      @attributes.dup
    end

    def to_json(opts={})
      as_json.to_json(opts)
    end

    def ==(other)
      case other
      when Essential::Resource
        self.sid == other.sid
      else
        false
      end
    end

    def inspect
      format(
        '#<%s:0x%s @attributes=%s>',
        self.class.name,
        (self.object_id << 1).to_s(16),
        JSON.pretty_generate(@attributes)
      )
    end

    protected
    def [](key)
      # realize attributes upon access
      self.fetch unless :sid == key.to_sym || loaded?
      @attributes[key]
    end

    def apply_schema(key, type)
      if val = @attributes[key]
        if type === val
          return
        elsif type.respond_to?(:parse)
          val = type.parse(@attributes[key])
          if type == Time && Essential.utc_offset
            # coerce into local time
            val = val.getlocal(Essential.utc_offset)
          end
          @attributes[key] = val
        else
          raise ArgumentError, format('unsure how to apply schema "%s"', type)
        end
      end
    end

  end
end
