module Essential
  class Resource
    module AttrRelations

      # Method for defining a mapping between a `sid` property and a Class.
      #
      def attr_relation(pairs)
        pairs.each do |method, resource|
          sid_method = method.to_sym

          # Enforces that we're dealing with a `sid` property.
          if method =~ /\A(.+)_sid\Z/
            method = $1.to_sym
          else
            raise ArgumentError, 'attribute must end in _sid'
          end

          unless method_defined?(sid_method)
            # perhaps we didn't explicitly define access to the sid - do so now.
            attr_property sid_method
          end

          # Avoid circular load dependencies:
          #
          # `resource` might be a Class, or it might be a String.
          #
          # If it is a String, we'll attempt to turn it into a CLass.
          # This will be done once - the first time it is accessed -
          # to allow both classes to load completely before constantizing
          # the String.
          my_lazy_resource = format('lazy_%s_resource', method).to_sym
          my_lazy_variable = format('@%s', my_lazy_resource).to_sym
          define_singleton_method(my_lazy_resource) do
            # only resolve this resource once
            lazy_resource = instance_variable_get(my_lazy_variable)
            unless lazy_resource
              lazy_resource = resource

              case lazy_resource
              when String
                if lazy_resource.include?('::')
                  # this appears to be a fully-namespaced Resource
                  lazy_resource = lazy_resource.split('::')
                else
                  # no namespace - assume it's part of the current class.
                  namespace = self.name.split('::')
                  namespace.pop
                  namespace << lazy_resource
                  lazy_resource = namespace
                end
                # fetch each module in turn
                lazy_resource = lazy_resource.reduce(Object) {|o,c| o.const_get c}
              end

              # now that we have a class, ensure it is *actually* a Resource!
              unless lazy_resource < Essential::Resource
                raise ArgumentError, format('mapping must be an %s', Essential::Resource.name)
              end

              # cache it.
              instance_variable_set(my_lazy_variable, lazy_resource)
            end

            lazy_resource
          end

          define_method(method) do
            sid = send(sid_method)
            if sid.nil?
              nil
            else
              self.class.send(my_lazy_resource).new(sid: sid, headers: @headers)
            end
          end
        end

      end

    end
  end
end
